/* Copyright (c) 1999 Regents of the University of California */

/******************************************************************************
 *
 *	UC Berkeley, 1999
 *	Haarm-Pieter Duiker		duiker@cs.berkeley.edu
 *	Tim Hawkins			tsh@cs.berkeley.edu
 *	Paul Debevec			debevec@cs.berkeley.edu
 *
 *	Implementing modified version of algorithm presented in
 *	Recovering High Dynamic Range Radiance Maps from Photographs
 *	Debevec, Malik 1997
 *
 *  mkhdr.c
 *	Take in a set of images with known exposure times 
 *      and produce a radiance map
 * 
 *  Output format: the RGBE standard of the Radiance rendering system
 *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include "tiffio.h"
#include "dkcolor.h"
#include "mkhdr.h"

#define MAX_NUM_IMAGES 128

int MIN = 0;
int MAX = 255;
#define RANGE 256
#define numBins 256
int BITDEPTH = 1;

int NUM_IMAGES = 0;
int POINTS = 0;

int INPUT_FILE = 0;
int OUTPUT_FILE = 0;
int INPUT_CURVE_FILE = 0;
int OUTPUT_CURVE_FILE = 0;

int HEIGHT = 0;
int WIDTH = 0;
int PIXELS = 0;
int CHANNELS = 1;

float TOTAL_WEIGHT = 0.f;

float LAMBDA1 = 20.f;
float LAMBDA2 = 1.f;

int VARIANCE = 0;
int DEBUG_OUTPUT = 0;
int TIFF_OUTPUT = 0;

float *gradWeights;
float *bellWeights;

#define min(x, y) (((x) < (y)) ? (x) : (y))
#define max(x, y) (((x) > (y)) ? (x) : (y))


void
processCommandLine(int argc, char** argv) {
  int i, limit;

  INPUT_FILE = argc-2;
  OUTPUT_FILE = argc-1;

  limit = argc;
  for(i=1;i<argc;i++) {
    if( strcmp("-h", argv[i]) == 0) {
      printf("usage:\t mkhdr [options] inputfile outputfile\n");
      printf("options:\t-c input_curve_file\n");
      printf("        \t-w ouput_curve_file\n\n");
    } else if( strcmp("-c", argv[i]) == 0) {
      INPUT_CURVE_FILE = ++i;
    } else if( strcmp( "-w", argv[i] ) == 0 ) {
      OUTPUT_CURVE_FILE = ++i;
    } else if( strcmp( "-l1", argv[i] ) == 0 ) {
      LAMBDA1 = atof(argv[++i]);
    } else if( strcmp( "-l2", argv[i] ) == 0 ) {
      LAMBDA2 = atof(argv[++i]);
    } else if( strcmp( "-var", argv[i] ) == 0 ) {
      VARIANCE = 1;
    } else if( strcmp( "-d", argv[i] ) == 0 ) {
      DEBUG_OUTPUT = 1;
    } else if( strcmp( "-max", argv[i] ) == 0 ) {
      /*      RANGE = atof(argv[++i])+1; */
    } else if( strcmp( "-fptiff", argv[i] ) == 0 ) {
      TIFF_OUTPUT = 1;
    } else {
      limit--;
    }
  }

  if(limit != (argc-2)) {
    printf("usage:\t mkhdr [options] inputfile outputfile\n");
    printf("options:\t-c input_curve_file\n");
    printf("        \t-w ouput_curve_file\n\n");
    exit(1);
  }
}

static int
compare(expImage* a, expImage* b) {
  return (((a->exposure - b->exposure) > 0) ? 1 : -1);
}

int
processInputFile(expImage* imageList, const char* filename) {
  int images;
  float shutter, fstop, gain, nd, smallest;
  char file[80];
	
  FILE* input;
  input = fopen(filename, "ro");

  if(input == NULL) {
    printf("mkhdr : Input file %s could not be opened. Operation aborted\n", filename);
    return 0;
  }
	
  smallest = HUGE_VAL;
  images = 0;
  while( !(fscanf(input, "%s %f %f %f %f\n", file, &shutter, &fstop, &gain, &nd) == EOF) &&
	 images<MAX_NUM_IMAGES ) {

    imageList[images].image = fopen(file, "ro");
    if(imageList[images].image != NULL) {

      imageList[images].exposure = -log(shutter * (fstop*fstop))
	+ log(512) + (gain/3.0) - nd;
      /*      imageList[images].exposure /= log(2); */

      imageList[images].data = processImageFile(imageList[images].image, file);

      if(imageList[images].exposure < smallest) 
	smallest = imageList[images].exposure;
      images++;

      printf("Image %s: shutter %6.3f fstop %6.3f gain %4.1f ND %4.1f \n\tExposure: %6.6f\n",file, shutter, fstop, gain, nd, imageList[images-1].exposure);	
    } else {
      printf("processingInputFile: File %s could not be opened.\n", file);
      return 0;  
    }	
  }

  /* Put the pictures into ascending order of exposure */
  qsort(imageList, images, sizeof(expImage), (void*)compare);

  NUM_IMAGES = images;
  POINTS = min((RANGE/NUM_IMAGES*4),RANGE);

  fclose(input);
  return 1;
}

static short**
processTIF(FILE* image, char* filename) {
  long ImageWidth =0, ImageLength =0, BitsPerSample =0, SamplesPerPixel =0;
  short **data;
  void *scanline;
  int i, j, x, y;
  TIFF* tiffImage;

  fclose(image);
  tiffImage = TIFFOpen(filename, "r");

  if(tiffImage) {
    TIFFGetField(tiffImage, TIFFTAG_IMAGEWIDTH, &ImageWidth);
    TIFFGetField(tiffImage, TIFFTAG_IMAGELENGTH, &ImageLength);
    TIFFGetField(tiffImage, TIFFTAG_SAMPLESPERPIXEL, &SamplesPerPixel);
    TIFFGetField(tiffImage, TIFFTAG_BITSPERSAMPLE, &BitsPerSample);

    /*
      printf("ImageWidth %d\n",  ImageWidth);
      printf("ImageLength %d\n",  ImageLength);
      printf("BitsPerSample %d\n",  BitsPerSample);
      printf("SamplesPerPixel %d\n",  SamplesPerPixel);
    */

    HEIGHT = ImageLength;
    WIDTH = ImageWidth;
    BITDEPTH = (BitsPerSample>>16)/8;
    CHANNELS = (SamplesPerPixel>>16);
    PIXELS = WIDTH*HEIGHT;
    
      
    data = (short**)calloc(sizeof(short*),CHANNELS);
    for(i=0;i<CHANNELS;i++) {
      data[i] = (short*)calloc(sizeof(short),PIXELS);
    }

    scanline = _TIFFmalloc(TIFFScanlineSize(tiffImage));
    _TIFFmemset(scanline, 0, TIFFScanlineSize(tiffImage));
    
    for(i=0;i<HEIGHT;i++) {
      TIFFReadScanline(tiffImage, scanline, i, 0);

      if (CHANNELS==1) {
	for(j=0;j<WIDTH;j++) {
	  data[0][j+WIDTH*i] = ((unsigned char*)scanline)[j];
	}
      } else if (CHANNELS == 3) {
	for (y=0;y<WIDTH;y++) {
	  for(x=0;x<CHANNELS;x++) {
	    data[x][y+WIDTH*i] = ((unsigned char*)scanline)[y*CHANNELS+x];
	  } 
	}
      }
    }
  
    _TIFFfree(scanline);
    TIFFClose(tiffImage);
    
    return data;
  } else {
    printf("mkhdr : Unable to open %s", filename);
    return 0;
  }
}

static short**
processPPM(FILE* image, char* filename) {
  int bpp;
  int i, x, y;
  char strbuf[100];
  
  short** data;

  strcpy(strbuf,"#");while (strbuf[0] == '#') fgets(strbuf,100,image);
  if (strncmp(strbuf,"P6",2)==0)   CHANNELS = 3;
  else if (strncmp(strbuf,"Pf",2)==0)   CHANNELS = 4; /* Floating point */
  else CHANNELS = 1;
  strcpy(strbuf,"#");while (strbuf[0] == '#') fgets(strbuf,100,image);
  sscanf(strbuf,"%d %d",&WIDTH,&HEIGHT);
  strcpy(strbuf,"#");while (strbuf[0] == '#') fgets(strbuf,100,image);

  PIXELS = WIDTH*HEIGHT;
	
  sscanf(strbuf,"%d",&bpp);
  if (bpp != 255) {
    BITDEPTH = 2;
  } else {
    BITDEPTH = 1;
  }

  printf("Size is %dx%d, %d channels, bit depth %d.\n",WIDTH,HEIGHT,CHANNELS,BITDEPTH);

  data = (short**)malloc(sizeof(short*)*CHANNELS);
  for(i=0;i<CHANNELS;i++) {
    data[i] = (short*)malloc(sizeof(short)*PIXELS);
  }


  if (CHANNELS==4) {
    printf("processImageFile: Floating point PPMs not supported\n");
    return 0;
  }

  if (CHANNELS==1) {
    fread(data[0],BITDEPTH,PIXELS,image);
  }
  else if (CHANNELS == 3) {
    for (y=0;y<PIXELS;y++) {
      for(x=0;x<CHANNELS;x++) {
	data[x][y] = fgetc(image);
      } 
    }
  }
  
  fclose(image);
  return data;
}

short**
processImageFile(FILE* image, char* filename) {
  if( strcmp( (filename+(strlen(filename)-3)), "ppm") == 0) {
    return processPPM(image, filename);
  } else if( strcmp( (filename+(strlen(filename)-3)), "tif") == 0) {
    return processTIF(image, filename);
  }
}

void
writeOutCurve(curveElement** curve, char* filename) {	
  int i,j;

  FILE* output;

  output = fopen(filename, "w+");
	
  if( output ) {
    fprintf(output, "C = [\n");
    for(i=0;i<RANGE;i++) {
      for(j=(CHANNELS-1);j>=0;j--) {
	fprintf(output, "%f ", curve[j][i]); 
      }
      fprintf(output, "\n");
    }
    fprintf(output, "];\n");
    fclose(output);
  } else {
    printf("mkhdr : Unable to open %s for output\n", filename);
  }
  printf(" Wrote out curve \n");
}

int
readInCurve(curveElement** curve, char* filename) {
  int i, j;
  float x;

  FILE* input;
  input = fopen(filename, "r");

  printf(" Reading in curve \n");
	
  if(input == NULL) {
    printf("mkhdr : Curve file %s could not be opened.\n", filename);
    printf("mkhdr : mkhdr will attempt to generate a new curve\n");
    return 0;
  }

  fscanf(input, "C = [\n");
  for(i=0;i<RANGE;i++) {
    for(j=(CHANNELS-1);j>=0;j--) {
      fscanf(input, "%f", &x); 
      curve[j][i] = x;
    }
    fscanf(input, "\n");
  }
  fscanf(input, "];\n");

  fclose(input);
  return 1;
}


int DKcompareFloats(const void *f1,const void *f2) {
	if (*(float*)f1<*(float*)f2) return -1;
	else return (*(float*)f1>*(float*)f2) ? 1 : 0;
}

#define DK_MEDIAN_SAMPLES 100

static float getApproxMedian(radianceMap* hdrImage) {
	long i,k;
	float vals[DK_MEDIAN_SAMPLES];
	float r,g,b,lum;

    for (k=0;k<DK_MEDIAN_SAMPLES;k++) {
		i = lrand48()%PIXELS;
		r = hdrImage->data[0][i];
		g = hdrImage->data[1][i];
		b = hdrImage->data[2][i];
		lum = .55f*g+.3f*r+.15f*b;  /* approximate luminance */
		vals[k] = lum;
	}
	qsort(vals,DK_MEDIAN_SAMPLES,sizeof(float),DKcompareFloats);/* int (*)(const void*,const void*)) */
	return vals[DK_MEDIAN_SAMPLES/2];
}

void
writeOutRadianceFile(radianceMap* radMap, char* filename) {
  FILE* theFileHandle;
  TIFF* theTiffFileHandle;
  
  int i, j, row;
  float exposure;
  char* comments = "no comment";

  /*Allocate enough space for one row of COLOR at a time */
  COLOR* oneRow;
  float* floatRow;

  if( TIFF_OUTPUT ) {
    floatRow = (float*)malloc(sizeof(float)*WIDTH*CHANNELS);

    theTiffFileHandle = TIFFOpen(filename, "w");
    
    if( theTiffFileHandle ) {      
      TIFFSetField(theTiffFileHandle, TIFFTAG_IMAGEWIDTH, WIDTH);
      TIFFSetField(theTiffFileHandle, TIFFTAG_IMAGELENGTH, HEIGHT);
      TIFFSetField(theTiffFileHandle, TIFFTAG_ROWSPERSTRIP, HEIGHT);
      TIFFSetField(theTiffFileHandle, TIFFTAG_SAMPLESPERPIXEL, CHANNELS);
      TIFFSetField(theTiffFileHandle, TIFFTAG_SAMPLEFORMAT, 3);
      TIFFSetField(theTiffFileHandle, TIFFTAG_BITSPERSAMPLE, 32);
      TIFFSetField(theTiffFileHandle, TIFFTAG_COMPRESSION, 1);
    
      TIFFSetField(theTiffFileHandle, TIFFTAG_PLANARCONFIG, 1);
      TIFFSetField(theTiffFileHandle, TIFFTAG_PHOTOMETRIC, 2);

      /* Convert separated channel representation to per pixel representation */
      for (row=0;row<HEIGHT;row++) { 
	for(i=0;i<WIDTH;i++) { 
	  for(j=0;j<CHANNELS;j++) { 
	    floatRow[i*CHANNELS+j] = radMap->data[j][i+row*WIDTH]; 
	  } 
	} 
	TIFFWriteRawStrip(theTiffFileHandle, 0, floatRow, sizeof(float)*WIDTH*CHANNELS); 
      } 
      TIFFClose(theTiffFileHandle);    
    } else {
      printf("mkhdr : Unable to open %s for output\n", filename);
    }

    free(floatRow);
  } else {
    oneRow = (COLOR*)malloc(sizeof(COLOR)*WIDTH);

    theFileHandle = fopen(filename, "wb");

    if( theFileHandle ) {
      exposure = 0.52/getApproxMedian(radMap);
  
      /*a very basic RADIANCE pic file header */
      fprintf(theFileHandle,"#?RADIANCE\n");
      fprintf(theFileHandle,"# %s\n",comments);
      fprintf(theFileHandle,"FORMAT=32-bit_rle_rgbe\n");
      fprintf(theFileHandle,"EXPOSURE=%25.13f\n",exposure);
      fprintf(theFileHandle,"\n");
      fprintf(theFileHandle,"-Y %d +X %d\n",HEIGHT, WIDTH);
    
      /* Convert separated channel representation to per pixel representation */
      for (row=0;row<HEIGHT;row++) { 
	for(i=0;i<WIDTH;i++) { 
	  for(j=0;j<CHANNELS;j++) { 
	    oneRow[i][j] = radMap->data[j][i+row*WIDTH]*exposure; 
	  } 
	} 
	fwritescan(oneRow, WIDTH, theFileHandle); 
      } 
      fclose(theFileHandle);
    } else {
      printf("mkhdr : Unable to open %s for output\n", filename);
    }
    free(oneRow);
  }

  printf(" Wrote out radiance file \n");
}

void
generateOneChannel(int channel, float* radMap, curveElement* curve, expImage* files) {
  int x, zv, i;
  double weightsum, irad, iradsum;

  for (x=0; x<PIXELS; x++) {
    weightsum=0;
    iradsum=0;
    for (i=0 ; i<NUM_IMAGES; i++) {
      zv = files[i].data[channel][x];
      iradsum += (curve[zv]-files[i].exposure) * 
	weight(x, files, channel, i) *
	bellWeight(zv);      
      weightsum +=  weight(x, files, channel, i) *
	bellWeight(zv);
    }
    if (weightsum>0) irad = iradsum / weightsum;
    else irad = -1000;
    radMap[x] = (float)(exp(irad));
  }
}

double
bellWeight(float i) {
  float ratio, w, subratio;
  /* The values either range over the given bit depth */
  /* or over the specified range */
  ratio = (i-MIN)/(MAX-MIN);
  ratio = min(max(ratio, 0), 1);

  ratio *= RANGE-1;
  subratio = (ratio-floor(ratio));
  w = (1-subratio)*bellWeights[(int) floor(ratio)] +
    subratio*bellWeights[(int) ceil(ratio)];
  return w;
}

double
weight(float zi, void* ptr, int channel, int image) {
  double w;
  int i, z1, z2, z3, weightRange;
  float *trace, fz1, fz2, fz3;
  expImage* files;
  
  z1 = 0; z2 = 0; z3 = 0;

  weightRange = RANGE-1;
  
  if( image == -1 ) {
    trace = (float*)ptr;
    i = (int)zi;

    fz2 = trace[i];

    if(zi == 0) {
      fz1 = 2*fz2 - trace[i+1];
    } else {
      fz1 = trace[i-1];
    }

    if (zi == (NUM_IMAGES-1)) {
      fz3 = 2*fz2 - trace[i-1];
    } else {
      fz3 = trace[i+1];
    }

    w = ((float)(fz3-fz1)/2)/weightRange;
    w = min(max(w, 0), 1);

    return w;
  } else {     
    files = (expImage*)ptr;
    i = (int) zi;

    z2 = files[image].data[channel][i];

    if(image == 0) {
      z1 = 2*z2 - files[image+1].data[channel][i];
    } else {
      z1 = files[image-1].data[channel][i];
    }

    if (image == (NUM_IMAGES-1)) {
      z3 = 2*z2 - files[image-1].data[channel][i];
    } else {
      z3 = files[image+1].data[channel][i];    
    }

    w = ((float)(z3-z1)/2)/weightRange;
    w = min(max(w, 0), 1);

    return w;
  }
}

static float
informationAdded(pixelTrace* pixelTraces, float* bins, int newTrace) {
  int i, j, lowVal, highVal;
  float currVal, addedInfo, newInfo, spread, maxHeight;

  addedInfo = 0;

  for(i=0;i<NUM_IMAGES;i++) {
    currVal = pixelTraces[newTrace].trace[i];
    spread = max(0.5, sqrt(currVal)/4);
    maxHeight = 1/spread;

    lowVal = max(floor(currVal-spread), 0);
    highVal = min(ceil(currVal+spread), (RANGE-1));

    for(j=lowVal;j<=highVal;j++) {
      newInfo = max(0, maxHeight*(1-(fabs(currVal-j)/spread)));
      addedInfo += (newInfo == 0) ? 0 : newInfo/(newInfo + bins[j]);
    }
  }  
  return addedInfo;  
}

static void
addInformation(pixelTrace* pixelTraces, float* bins, int newTrace) {
  int i, j, lowVal, highVal;
  float currVal, spread, maxHeight;

  for(i=0;i<NUM_IMAGES;i++) {
    currVal = pixelTraces[newTrace].trace[i];
    spread = max(0.5, sqrt(currVal)/4);
    maxHeight = 1/spread;

    lowVal = max(floor(currVal-spread), 0);
    highVal = min(ceil(currVal+spread), (RANGE-1));

    for(j=lowVal;j<=highVal;j++) {
      bins[ j ] += max(0, maxHeight*(1-(fabs(currVal-j)/spread)));
    }    
  }
}

static int
pickPoint(float* bins, int nmBins, pixelTrace* pixelTraces) {
  int i, maxTrace;
  float tempWeight, maxWeight;

  maxWeight = 0;
  maxTrace = 0;
  for(i=0;i<RANGE*NUM_IMAGES;i++) {
    if( !pixelTraces[i].selected && (pixelTraces[i].trace != NULL) ) {
      tempWeight = informationAdded(pixelTraces, bins, i);
      tempWeight *= pixelTraces[i].weight;
      if( tempWeight > maxWeight) {
	maxTrace = i;
	maxWeight = tempWeight;
      }	
    }
  }
 
 addInformation(pixelTraces, bins, maxTrace);
 pixelTraces[maxTrace].selected = 1;

 /*printf("%d, %d\t%3.2f\t%1.4f\t%3.2f\n", maxTrace, maxTrace%RANGE, maxWeight/pixelTraces[maxTrace].weight, pixelTraces[maxTrace].weight, maxWeight); */
  
  return maxTrace;
}

static float
gradientWeight(int Zi, expImage* files, int channel) {
  int i, j, image, Zij;
  float sum, currWeight;

  if(gradWeights[Zi] > 0) {
    return gradWeights[Zi];
  } else {

    currWeight = 0;
    for(i=0;i<NUM_IMAGES;i++) {
      if(bellWeight(files[i].data[channel][Zi]) >= currWeight) {
	image = i;
	Zij = files[image].data[channel][Zi];
	currWeight = bellWeight(Zij);
      }
    } 
  
    sum = 0;
    for(i=max(Zi-1,0);i<=min(Zi+1,PIXELS);i++) {
      for(j=((i-WIDTH)>0 ? -1 : 0);j<=((i+WIDTH)>PIXELS ? 0 : 1);j++) {
	sum += fabs(Zij-files[image].data[channel][i+WIDTH*j]);
      }
    }
    sum /= (8*Zij);
  
    gradWeights[Zi] = 1/(sum+1);
    return 1/(sum+1);
  }
}

static void
computeVariance(pixelTrace* pixelTraces, expImage* files,
		int channel, int* valid) {
  float *variance, *varianceWeights;
  int i, j, Zij, k, t1;
  float varianceExpVal, varianceExpValWeight, currVariance, t2;
  
  
  /* Variance calculations */
  variance = (float*)calloc(sizeof(float), NUM_IMAGES*RANGE);
  varianceWeights = (float*)calloc(sizeof(float), NUM_IMAGES*RANGE);
  
  for(i=0;i<NUM_IMAGES;i++) {
    for(j=0;j<PIXELS;j++) {
      currVariance = 0;
      Zij = files[ i ].data[channel][j];
      if( valid[ j ] &&
	  (pixelTraces[ Zij + RANGE*i].trace != NULL) ) {
	for(k=0;k<NUM_IMAGES;k++) {
	  t1 = files[ k ].data[channel][j] -
	    pixelTraces[ Zij + RANGE*i ].trace[k];
	  currVariance += t1*t1;

	}	

	t2 = gradientWeight(j, files, channel)*
	  bellWeight(Zij);
	variance[ Zij + RANGE*i ] += t2*sqrt(currVariance);
	varianceWeights[ Zij + RANGE*i ] += t2;
      }
    }
  }
  for(i=0;i<NUM_IMAGES*RANGE;i++) {
    if(varianceWeights[ i ] > 0) {
      variance[ i ] /= varianceWeights[ i ];
    }
  }

  varianceExpVal = 0;
  varianceExpValWeight = 0;
  for(i=0;i<NUM_IMAGES*RANGE;i++) {
    varianceExpVal += variance[i]*bellWeight(i%RANGE);
    varianceExpValWeight += bellWeight(i%RANGE);
  }
  varianceExpVal /= varianceExpValWeight;

  for(i=0;i<NUM_IMAGES*RANGE;i++) {
    pixelTraces[i].weight *= 1/(1 + variance[i]/varianceExpVal);
  }

  free(variance);
  free(varianceWeights);
}

static void
computeMinMax(int *totHist) {
  int i, j, newMin, newMax;
  float percentage, stretch;

  percentage = 0.05;
  stretch = (MAX-MIN)*percentage;
  
  newMin = MIN;
  for(i=MIN+1;i<stretch+MIN;i++) {
    if(totHist[i] >=
       totHist[newMin]* (0.5+0.5*((i-MIN)/stretch)) ) {
       newMin = i;
    }
  }	

  newMax = MAX;
  for(i=MAX-1;i>MAX-stretch;i--) {
    if(totHist[i] >=
       totHist[newMax]* (0.5+0.5*(-(i-MAX)/stretch)) ) {
       newMax = i;
    }
  }	
  /*
  printf("MIN: %d, newMin: %d\n", MIN, newMin);
  printf("MAX: %d, newMAX: %d\n", MAX, newMax);

  for(i=MIN;i<=newMin+1;i++) {
    printf("%d ", totHist[i]);
  }
  printf("\n");

  for(i=MAX;i>=newMax-1;i--) {
    printf("%d ", totHist[i]);
  }
  printf("\n");
  */
  MIN = newMin;
  MAX = newMax;
}

static void
findAveragePixelTraces(pixelTrace *pixelTraces, expImage* files, int channel) {
  int i, j, k, Zi, Zij;
  int *valid;
  float newWeight, maxWeight, currVariance, varianceExpVal, varianceExpValWeight;
  float *variance, *varianceWeights;

  /* TESTING */
  int maxHist[RANGE], minHist[RANGE], totHist[RANGE];
  FILE* hist;
  char str[10];
  /* TESTING */
  
  for(i=0;i<RANGE;i++) {
    maxHist[i] = 0;
    minHist[i] = 0;
    totHist[i] = 0;
  }

  valid = (int*)calloc(sizeof(int),PIXELS);  
  /* Invalid pixel. Disregard completely */
  for(i=0;i<PIXELS;i++) {
    valid[ i ] = 1;
    for(k=0;k<NUM_IMAGES;k++) {
      if( files[ k ].data[ 0 ][ i ] == 0 &&
	  files[ k ].data[ 1 ][ i ] == 0 &&
	  files[ k ].data[ 2 ][ i ] == 255 ) {
	valid[ i ] *= 0;
      }
    }
  }

  fflush(stdout);

  maxWeight = 0;  
  for(i=0;i<NUM_IMAGES;i++) {
    for(Zi=0;Zi<PIXELS;Zi++) {
      if( valid[ Zi ] ) {
	Zij = files[ i ].data[ channel ][ Zi ];
	if( pixelTraces[ Zij+RANGE*i ].trace == NULL ) {
	  pixelTraces[ Zij+RANGE*i ].trace = 
	    (float*)calloc( sizeof(float),NUM_IMAGES );
	  pixelTraces[ Zij+RANGE*i ].weight = 0;
	  pixelTraces[ Zij+RANGE*i ].image = i;
	  pixelTraces[ Zij+RANGE*i ].value = Zij;
	}
      
	newWeight = gradientWeight(Zi, files, channel);
	pixelTraces[Zij+RANGE*i].weight += newWeight;
	for(k=0;k<NUM_IMAGES;k++) {
	  pixelTraces[Zij+RANGE*i].trace[k] += 
	    files[k].data[channel][Zi]*newWeight;
	}      
      }
    }
      
    for(Zij=0;Zij<RANGE;Zij++) {
      for(k=0;k<NUM_IMAGES;k++) {
	if(pixelTraces[Zij+RANGE*i].trace != NULL) {
	  pixelTraces[Zij+RANGE*i].trace[k] /= pixelTraces[Zij+RANGE*i].weight;
	  
	  if(pixelTraces[Zij+RANGE*i].trace[k] > (RANGE-1)) {
	    pixelTraces[Zij+RANGE*i].trace[k] = (RANGE-1);
	  }
	}	  
      }

      pixelTraces[Zij+RANGE*i].weight =
	sqrt(pixelTraces[Zij+RANGE*i].weight);

      pixelTraces[Zij+RANGE*i].weight *= bellWeight(Zij);
      
      if(  pixelTraces[Zij+RANGE*i].weight > maxWeight ) {
	maxWeight = pixelTraces[Zij+RANGE*i].weight;
      }
    }
    
    printf("%d ", (i+1));
    fflush(stdout); 
  }
  printf("\n");  

  for(i=0;i<NUM_IMAGES*RANGE;i++) {
    if(pixelTraces[i].trace != NULL) {
      pixelTraces[i].weight /= maxWeight;
      
      minHist[ (int) pixelTraces[i].trace[0] ] += 1;
      maxHist[ (int) pixelTraces[i].trace[NUM_IMAGES-1] ] += 1;
      for(j=0;j<NUM_IMAGES;j++) {
	totHist[ (int) pixelTraces[i].trace[j] ] += 1;
      }
    }
  }

  for(i=0;i<RANGE;i++) {
    if(minHist[i] > minHist[MIN]) {
      MIN = i;
    }
    if(maxHist[i] > maxHist[MAX]) {
      MAX = i;
    }
  }
  
  computeMinMax(totHist);

  if( VARIANCE ) {
    i = time(NULL);
    computeVariance(pixelTraces, files, channel, valid);
    printf("variance took %d seconds\n", time(NULL) - i);
  }
  
  free(valid);
}

/* Effectively hard coded to work with 8 bits right now. */
static void
generateCurveLogLog(int channel, curveElement* curve, expImage* files) {
  int i, j, k, val, Zi, Zij; 
  float l, wij, sum, Pij, Eij, ratio;
  char str[10];
  
  /* LAPACK */
  int M, N, NRHS = 1, LDA, LDB, RANK, INFO, LWORK;
  double RCOND = -1;
  double *A, *B, *oldB, *S, *WORK;
  /* LAPACK */

  int Ah = (POINTS*NUM_IMAGES+RANGE+1);
  int Aw = (POINTS+RANGE);

  float bins[numBins];
  float binRange;
  int *pts;
  pixelTrace *pixelTraces;
  FILE* points;

  /* TESTING */
  int hist[RANGE/16];
  pts = calloc(sizeof(int),2*POINTS*NUM_IMAGES);

  for(i=0;i<numBins;i++) { bins[i] = 0; }
  for(i=0;i<RANGE/16;i++) { hist[i] = 0; }
  /* TESTING */

  l = LAMBDA1;
  
  /* LAPACK */
  M = LDA = LDB = Ah;
  N = Aw;  
  LWORK = 15*(3*min(M,N) + max(2*min(M,N), max(M,N)));

  S = (double*)calloc(sizeof(double),min(M,N)); 
  WORK = (double*)calloc(sizeof(double),LWORK); 
  A = (double*)calloc(sizeof(double),M*N);
  B = (double*)calloc(sizeof(double),M);
  /* LAPACK */

  printf("Generating averages of pixel traces... ");
  pixelTraces = (pixelTrace*)calloc(sizeof(pixelTrace),RANGE*NUM_IMAGES);
  findAveragePixelTraces(pixelTraces, files, channel);

  printf("Generating curve for channel %d ...\n", channel+1);
  printf(" %d points in the range [%d, %d], A is %dx%d, B is %dx1\n",
	 POINTS,MIN,MAX,Ah,Aw,Ah);

  /* INITIALIZATION */
  binRange = log(MAX-MIN+1)/(numBins-1);
  
  sum = 0;
  k = 0;
  for(i=0;i<POINTS;i++) {
    Zi = pickPoint(bins, numBins, pixelTraces);

    pts[i] = Zi;    

    for(j=0;j<NUM_IMAGES;j++) {
      Zij = max((min(pixelTraces[Zi].trace[j], MAX) -MIN), 0) + 1;
      Pij = log(Zij);
    
      wij = weight(j,pixelTraces[Zi].trace,0,-1)*
	bellWeight(pixelTraces[Zi].trace[j]);

      sum += wij*wij;
      ratio = (Pij/binRange-floor(Pij/binRange));

      A[k + (Ah*((int)floor(Pij/binRange)))] = wij*(1-ratio);
      A[k + (Ah*((int)ceil(Pij/binRange)))] = wij*ratio;
      A[k + (Ah*(RANGE+i))] = -wij;
      B[k] = wij*files[j].exposure;

      hist[(int) floor(pixelTraces[Zi].trace[j])/(RANGE/16)] += 1;
      k++;
    }
  }

  A[k+ Ah*((int)floor(log(RANGE/2)/binRange))] = 1;
  k++;

  sum = sqrt(sum/TOTAL_WEIGHT);
  ratio = (float)(RANGE-1)/(MAX-MIN);

  for(i=0;i<(RANGE-3);i++) {
    wij = bellWeight(i/ratio+1+MIN)*sum;
    A[k+ (Ah*i)] = l*wij;
    A[k+ (Ah*(i+1))] = -2*l*wij;
    A[k+ (Ah*(i+2))] = l*wij;
    k++;
  }
  /* INITIALIZATION */

  /* TESTING */
  printf("Calling LAPACK with the following distribution\n ");
  for(i=0;i<(RANGE/16);i++) {
    printf("%3d ", hist[i]);
  }
  printf("\n");
  /* TESTING */

  /* LAPACK */
  i = time(NULL);
  dgelss_(&M, &N, &NRHS, A, &LDA, B, &LDB, S, &RCOND, &RANK, WORK, &LWORK, &INFO);
  printf("LAPACK took %d seconds.\n", time(NULL)-i);

  if( INFO < 0 ) {
    printf("The %d-th argument had an illegal value.\n", INFO);
  } else if ( INFO > 0 ) {
    printf("the algorithm for computing the SVD failed to converge.\n");
  }
  /* LAPACK */

  /* Store information */
  for(i=MIN;i<=MAX;i++) {
    val = i+1-MIN;
    ratio = log(val)/binRange-floor(log(val)/binRange);
    curve[i] = (1-ratio)*
      B[(int)floor(log(val)/binRange)]
      +
      ratio*
      B[(int)ceil(log(val)/binRange)];
  }
  for(i=0;i<MIN;i++) {
    curve[i] = curve[MIN];
  }
  for(i=MAX;i<RANGE;i++) {
    curve[i] = curve[MAX];
  }
  
  /* TESTING */
  printf("LAMBDA: %f, RANK: %d\n", LAMBDA1, RANK);

  if( DEBUG_OUTPUT ) {

    /*
    sprintf(str, "curve%d.m", channel);
    points = fopen(str, "w");
    fprintf(points, "C%d = [\n", channel);
    for(i=0;i<RANGE;i++) { 
      fprintf(points, "%f\n", curve[i]);
    }
    fprintf(points,"];\n");
    fclose(points);
    */
    
    sprintf(str, "points%d.m", channel);
    points = fopen(str, "w");
    fprintf(points, "P%d = [\n", channel);
    for(i=0;i<POINTS;i++) { 
      for(j=0;j<NUM_IMAGES;j++) {
	Zij = max((min(pixelTraces[ pts[i] ].trace[j], MAX) -MIN), 0) + 1;
	fprintf(points, "%f %f\n",
		log(Zij)/binRange+1,
		B[RANGE+i]+files[j].exposure);
      }
      fprintf(points, "\n");
    } 
    fprintf(points,"];\n");
    fclose(points);     

    /*
    sprintf(str, "B1%d.m", channel);
    points = fopen(str, "w");
    sprintf(str, "B%dc1", channel);
    fprintf(points, "%s = [\n", str);
    for(i=0;i<RANGE;i++) {
      fprintf(points, "%f\n", B[i]);
    }
    fprintf(points, "];\n");  
    fclose(points);
    */
  }
  /* TESTING */
  /* First call to LAPACK is done */


  /* LAPACK 2 */
  /* Data fitting call will now begin */
  printf("Beginning second stage data fitting\n");
  /* LAPACK */
  for(i=0;i<Ah*Aw;i++) {
    A[i] = 0;
  }

  Ah = POINTS*NUM_IMAGES+RANGE+1;
  Aw = RANGE;
  M = LDA = LDB = Ah;
  N = Aw;  
  LWORK = 15*(3*min(M,N) + max(2*min(M,N), max(M,N)));
  oldB = B;
  B = (double*)calloc(sizeof(double),M);
  /* LAPACK */

  /* INITIALIZATION */
  l = LAMBDA2;
  
  sum = 0;
  k = 0;
  for(i=0;i<POINTS;i++) {
    Zi = pts[i];
    for(j=0;j<NUM_IMAGES;j++) {
      Zij = max((min(pixelTraces[Zi].trace[j], MAX) -MIN), 0) + 1;
      Pij = log(Zij);
      wij = weight(j,pixelTraces[Zi].trace,0,-1)*
	bellWeight(pixelTraces[Zi].trace[j]);

      sum += wij*wij;
      ratio = (Pij/binRange-floor(Pij/binRange));

      A[k + (Ah*((int)floor(Pij/binRange)))] = wij*(1-ratio);
      A[k + (Ah*((int)ceil(Pij/binRange)))] = wij*ratio;
      
      Eij = (oldB[RANGE+i]+files[j].exposure);
      B[k] = wij*Eij;
      k++;
    }
  }

  A[k+ Ah*((int)floor(log(RANGE/2)/binRange))] = 1;
  k++;

  sum = sqrt(sum/TOTAL_WEIGHT);
  ratio = (float)(RANGE-1)/(MAX-MIN);

  for(i=0;i<(RANGE-3);i++) {
    wij = bellWeight(i/ratio+1+MIN)*sum;
    A[k+ (Ah*i)] = l*wij;
    A[k+ (Ah*(i+1))] = -2*l*wij;
    A[k+ (Ah*(i+2))] = l*wij;
    k++;
  }
  /* INITIALIZATION */

  /* LAPACK */
  i = time(NULL);
  dgelss_(&M, &N, &NRHS, A, &LDA, B, &LDB, S, &RCOND, &RANK, WORK, &LWORK, &INFO);
  printf("LAPACK took %d seconds, the second time\n", time(NULL)-i);

  if( INFO < 0 ) {
    printf("The %d-th argument had an illegal value.\n", INFO);
  } else if ( INFO > 0 ) {
    printf("the algorithm for computing the SVD failed to converge.\n");
  }
  /* LAPACK */

  /* Store information */
  for(i=MIN;i<=MAX;i++) {
    val = i+1-MIN;
    ratio = log(val)/binRange-floor(log(val)/binRange);
    curve[i] = (1-ratio)*
      B[(int)floor(log(val)/binRange)]
      +
      ratio*
      B[(int)ceil(log(val)/binRange)];
  }
  for(i=0;i<MIN;i++) {
    curve[i] = curve[MIN];
  }
  for(i=MAX;i<RANGE;i++) {
    curve[i] = curve[MAX];
  }

  if( DEBUG_OUTPUT ) {
    sprintf(str, "B%d.m", channel);
    points = fopen(str, "w");
    sprintf(str, "B%dc", channel);
    fprintf(points, "%s = [\n", str);
    for(i=0;i<RANGE;i++) {
      fprintf(points, "%f\n", B[i]);
    }
    fprintf(points, "];\n");  
    fclose(points);
  }
  /* LAPACK 2 */
  
  for(i=0;i<RANGE*NUM_IMAGES;i++) {
    if( pixelTraces[i].trace != 0 ) {
      free(pixelTraces[i].trace);
    }
  }
  free(pixelTraces);

  free(A);
  free(oldB);
  free(B);
  free(S);
  free(WORK);

  printf("Curve %d generated.\n",(channel+1));
}

void
generateCurve(int channel, curveElement* curve, expImage* files) {
  generateCurveLogLog(channel, curve, files);
}

void
createRadMapAndCurve(radianceMap* radMap, curveElement** curve, expImage* files) {
  int i, j;
		
  printf(" Initializing curve \n");
  for(i=0;i<CHANNELS;i++) {
    for(j=0;j<RANGE;j++) {
      curve[i][j] = 0;
    }
  }

  radMap->data = (float**)malloc(sizeof(float**)*CHANNELS);
  for(i=0;i<CHANNELS;i++) {
    radMap->data[i] = (float*)malloc(sizeof(float)*PIXELS);
    for(j=0;j<PIXELS;j++) {
      radMap->data[i][j] = 0;
    }
  }


  printf(" Generate curve and radiance map for each channel \n");
  for(i=0;i<CHANNELS;i++) {

    gradWeights = (float*)calloc(sizeof(float), PIXELS);
    j = time(NULL);
    
    generateCurve(i, curve[i], files);
    generateOneChannel(i,radMap->data[i],curve[i],files);

    printf("\nCurve %d generation took %d seconds\n", i, time(NULL)-j);
    free(gradWeights);
  }
}

void
createRadMapFromCurve(radianceMap* radMap, curveElement** curve, expImage* files) {
  int i, j;

  /* Allocate memory and Initialize radiance map */
  radMap->data = (float**)malloc(sizeof(float*)*CHANNELS);		       

  for(i=0;i<CHANNELS;i++) {
    radMap->data[i] = (float*)malloc(sizeof(float)*PIXELS);
    for(j=0;j<PIXELS;j++) {
      radMap->data[i][j] = 0;
    }
  }

  for(i=0;i<CHANNELS;i++) {
    printf(" Generating radiance map for channel %d\n", i);
    generateOneChannel(i,radMap->data[i],curve[i],files);
  }

  return;
}

void
precomputeBellWeights(float* weights) {
  float w, ratio;
  int i;

  for(i=0;i<RANGE;i++) {
    ratio = (float)i/(RANGE-1);
    w = (1-4*(ratio-0.5)*(ratio-0.5));
    weights[i] = w*w;
  }
}

int
main(int argc, char** argv) {

  int i;
  
  radianceMap *radMap;
  curveElement **curve;
  expImage* files;

  files = (expImage*)malloc(sizeof(expImage)*MAX_NUM_IMAGES);
  radMap = (radianceMap*)malloc(sizeof(radianceMap));

  bellWeights = (float*)calloc(sizeof(float),RANGE);
  precomputeBellWeights(bellWeights);
  
  srand48(time(NULL));
  
  processCommandLine(argc, argv);

  /*   Set globals HEIGHT, WIDTH, PIXELS, CHANNELS, and NUM_IMAGES */
  if( processInputFile(files, argv[INPUT_FILE]) ) {

    curve = (curveElement**)malloc(sizeof(curveElement*)*CHANNELS);
    for(i=0;i<CHANNELS;i++) {
      curve[i] = (curveElement*)malloc(sizeof(curveElement)*RANGE);
    }

    if( INPUT_CURVE_FILE &&
	readInCurve(curve, argv[INPUT_CURVE_FILE]) ) {

      createRadMapFromCurve(radMap, curve, files);

      if( OUTPUT_CURVE_FILE ) { 
	writeOutCurve(curve, argv[OUTPUT_CURVE_FILE]); 
      }
      writeOutRadianceFile(radMap, argv[OUTPUT_FILE]);
    } else if( !INPUT_CURVE_FILE ){
      /* Generated sum of weights over range for later normalization */
      for(i=0;i<RANGE;i++){
	TOTAL_WEIGHT += (bellWeight(i)*
			 bellWeight(i));
      }

      createRadMapAndCurve(radMap, curve, files);

      if( OUTPUT_CURVE_FILE ) { 
	writeOutCurve(curve, argv[OUTPUT_CURVE_FILE]); 
      }
      writeOutRadianceFile(radMap, argv[OUTPUT_FILE]);
    }

    for(i=0;i<CHANNELS;i++) { free( radMap->data[i] ); }
    free(radMap->data);
    for(i=0;i<CHANNELS;i++) { free( curve[i] ); }
    free(curve);
    for(i=0;i<NUM_IMAGES;i++) { free( files[i].data ); }
  }

  free(bellWeights);
  free(radMap);
  free(files);
}





